

/**********************************
* This generator transforms a given static CARP instance on an undirected graph to a set of corresponding uncertain CARP instances.
* The random variables in the uncertain version include the presence and demands of the required edges,
* and the presence and traversing cost of the non-required edges.
* For each required edge, it is required to be served with a probability of REQ_PROB, which is set to 0.9 here.
* If it is required, its demand is Gamma distributed with a probability density function of G(A, B), in which A = 20 and B = D/A, where D is the static demand.
* For each non-required edge, it is linked with a probability of LINK_PROB, which is set to 0.95 here.
* If it appears, its traversing cost is Gamma distributed with a probability density function of G(A, B), in which A = 20 and B = TC/A, where TC is the static traversing cost
**********************************/


#include <iostream>
#include <fstream>
#include <sstream>
#include <direct.h>
#include <math.h>
#include <vector>
#include <string>
#include <time.h>
#include <float.h>
#include <boost/multi_array.hpp>

#include "randomc.h"
#include "link.h"
#include "dijkstra.h"

using namespace std;
using namespace boost;

#define UNCERTAIN_INST_NUM 30

#define REQ_PROB 0.9
#define LINK_PROB 0.95
#define DEMMAND_SHAPE 20.0
#define TCOST_SHAPE 20.0

//string instance_name("egl-e1-A");

int vertex_num;
int req_edge_num;
int nonreq_edge_num;
int vehicle_num;
int capacity;

int DEPOT;

int main(void)
{
	vector<string> instance_names;
	instance_names.push_back("gdb1");
	instance_names.push_back("gdb2");
	instance_names.push_back("gdb3");
	instance_names.push_back("gdb4");
	instance_names.push_back("gdb5");
	instance_names.push_back("gdb6");
	instance_names.push_back("gdb7");
	instance_names.push_back("gdb8");
	instance_names.push_back("gdb9");
	instance_names.push_back("gdb10");
	instance_names.push_back("gdb11");
	instance_names.push_back("gdb12");
	instance_names.push_back("gdb13");
	instance_names.push_back("gdb14");
	instance_names.push_back("gdb15");
	instance_names.push_back("gdb16");
	instance_names.push_back("gdb17");
	instance_names.push_back("gdb18");
	instance_names.push_back("gdb19");
	instance_names.push_back("gdb20");
	instance_names.push_back("gdb21");
	instance_names.push_back("gdb22");
	instance_names.push_back("gdb23");
	instance_names.push_back("val1A");
	instance_names.push_back("val1B");
	instance_names.push_back("val1C");
	instance_names.push_back("val2A");
	instance_names.push_back("val2B");
	instance_names.push_back("val2C");
	instance_names.push_back("val3A");
	instance_names.push_back("val3B");
	instance_names.push_back("val3C");
	instance_names.push_back("val4A");
	instance_names.push_back("val4B");
	instance_names.push_back("val4C");
	instance_names.push_back("val4D");
	instance_names.push_back("val5A");
	instance_names.push_back("val5B");
	instance_names.push_back("val5C");
	instance_names.push_back("val5D");
	instance_names.push_back("val6A");
	instance_names.push_back("val6B");
	instance_names.push_back("val6C");
	instance_names.push_back("val7A");
	instance_names.push_back("val7B");
	instance_names.push_back("val7C");
	instance_names.push_back("val8A");
	instance_names.push_back("val8B");
	instance_names.push_back("val8C");
	instance_names.push_back("val9A");
	instance_names.push_back("val9B");
	instance_names.push_back("val9C");
	instance_names.push_back("val9D");
	instance_names.push_back("val10A");
	instance_names.push_back("val10B");
	instance_names.push_back("val10C");
	instance_names.push_back("val10D");
	instance_names.push_back("egl-e1-A");
	instance_names.push_back("egl-e1-B");
	instance_names.push_back("egl-e1-C");
	instance_names.push_back("egl-e2-A");
	instance_names.push_back("egl-e2-B");
	instance_names.push_back("egl-e2-C");
	instance_names.push_back("egl-e3-A");
	instance_names.push_back("egl-e3-B");
	instance_names.push_back("egl-e3-C");
	instance_names.push_back("egl-e4-A");
	instance_names.push_back("egl-e4-B");
	instance_names.push_back("egl-e4-C");
	instance_names.push_back("egl-s1-A");
	instance_names.push_back("egl-s1-B");
	instance_names.push_back("egl-s1-C");
	instance_names.push_back("egl-s2-A");
	instance_names.push_back("egl-s2-B");
	instance_names.push_back("egl-s2-C");
	instance_names.push_back("egl-s3-A");
	instance_names.push_back("egl-s3-B");
	instance_names.push_back("egl-s3-C");
	instance_names.push_back("egl-s4-A");
	instance_names.push_back("egl-s4-B");
	instance_names.push_back("egl-s4-C");

	for (vector<string>::iterator inst = instance_names.begin(); inst != instance_names.end(); inst++)
	{
		string instance_name(*inst);

	// get the static instance file

	string static_instance = "StaticInstances/"+instance_name+".dat";

	// name the output uncertain instances

	vector<string> uncertain_instance;
	for (int i = 0; i < UNCERTAIN_INST_NUM; i++)
	{
		string dir = "UncertainInstances/"+instance_name;
		mkdir(dir.c_str());

		stringstream ss;
		ss << i+1;
		string tmp_uinst = "UncertainInstances/"+instance_name+"/"+ss.str()+".dat";
		uncertain_instance.push_back(tmp_uinst);
	}

	// generate random seed;

	int seed = 0;
	CRandomMersenne RanGen(seed);

	vector<link> req_edge, nonreq_edge;

	// get the information of the static file

	ifstream infile(static_instance.c_str());

	string word;

	while (1)
	{
		infile >> word;

		if (word == "VERTICES")
		{
			infile >> word >> vertex_num;
		}
		else if (word == "ARISTAS_REQ")
		{
			infile >> word >> req_edge_num;
		}
		else if (word == "ARISTAS_NOREQ")
		{
			infile >> word >> nonreq_edge_num;
		}
		else if (word == "LISTA_ARISTAS_REQ")
		{
			infile >> word >> word;

			int hv, tv;
			double tc, sc, de;

			string line;
			for (int i = 1; i <= req_edge_num; i++)
			{
				getline(infile, line);
				string val_str;
				for (string::iterator iter = line.begin(); iter != line.end(); iter++)
				{
					if (isdigit(*iter) || isspace(*iter))
					{
						val_str = val_str+(*iter);
					}
				}
				istringstream digit_stream(val_str);
				digit_stream >> hv >> tv >> tc >> de;
				sc = tc;
				link tmp_link(hv, tv, tc, sc, de, 0);
				req_edge.push_back(tmp_link);
			}
		}
		else if (word == "LISTA_ARISTAS_NOREQ")
		{
			infile >> word >> word;

			int hv, tv;
			double tc;

			string line;
			for (int i = 1; i <= nonreq_edge_num; i++)
			{
				getline(infile, line);
				string val_str;
				for (string::iterator iter = line.begin(); iter != line.end(); iter++)
				{
					if (isdigit(*iter) || isspace(*iter))
					{
						val_str = val_str+(*iter);
					}
				}
				istringstream digit_stream(val_str);
				digit_stream >> hv >> tv >> tc;
				link tmp_link(hv, tv, tc, 0, 0, 0);
				nonreq_edge.push_back(tmp_link);
			}
		}
		else if (word == "DEPOSITO")
		{
			infile >> word >> DEPOT;
			break;
		}
	}

	infile.close();

	for (vector<string>::iterator iter = uncertain_instance.begin(); iter != uncertain_instance.end(); iter++)
	{
		vector<link> this_req_edge;
		vector<link> this_nonreq_edge;
		int this_req_edge_num;
		int this_nonreq_edge_num;
		double total_serv_cost;

		int valid = 0;
		while (!valid)
		{
			this_req_edge = req_edge;
			this_nonreq_edge = nonreq_edge;
			this_req_edge_num = req_edge_num;
			this_nonreq_edge_num = nonreq_edge_num;
			total_serv_cost = 0;

			// sample the random variables

			for (vector<link>::iterator iter2 = this_req_edge.begin(); iter2 != this_req_edge.end(); iter2++)
			{
				if (RanGen.Random() < REQ_PROB) // this edge is required
				{
					// sample its demand

					iter2->set_demand(RanGen.GamRandom(DEMMAND_SHAPE, iter2->get_demand()/(REQ_PROB*DEMMAND_SHAPE)));
				}
				else // this edge is not required
				{
					iter2->set_demand(0);
				}

				if (iter2->get_demand() > 0 || RanGen.Random() < LINK_PROB) // if this edge is required, then it is necessarily linked
				{
					// sample its traversing cost

					iter2->set_trav_cost(RanGen.GamRandom(TCOST_SHAPE, iter2->get_trav_cost()/TCOST_SHAPE));
				}
				else
				{
					iter2->set_trav_cost(DBL_MAX);
				}
			}
			
			for (vector<link>::iterator iter2 = this_nonreq_edge.begin(); iter2 != this_nonreq_edge.end(); iter2++)
			{
				if (RanGen.Random() < LINK_PROB) // if this edge is linked
				{
					// sample its traversing cost

					iter2->set_trav_cost(RanGen.GamRandom(TCOST_SHAPE, iter2->get_trav_cost()/TCOST_SHAPE));
				}
				else
				{
					iter2->set_trav_cost(DBL_MAX);
				}
			}

			// update this_req_edge_num and this_nonreq_edge_num;

			for (vector<link>::iterator iter2 = this_req_edge.begin(); iter2 != this_req_edge.end(); iter2++)
			{
				if (iter2->get_demand() == 0)  // this required edge is no longer required
				{
					this_req_edge_num --;
					this_nonreq_edge_num ++;

					if (iter2->get_trav_cost() == DBL_MAX) // this required edge is neither required nor linked
						this_nonreq_edge_num --;
				}
			}

			for (vector<link>::iterator iter2 = this_nonreq_edge.begin(); iter2 != this_nonreq_edge.end(); iter2++)
			{
				if (iter2->get_trav_cost() == DBL_MAX) // this non-required edge is no longer linked
					this_nonreq_edge_num --;
			}

			//// remove the non-required edges that are not linked
			//
			//vector<link>::size_type step = 0;
			//vector<link>::iterator iter3 = this_nonreq_edge.begin()+step;
			//while (iter3 != this_nonreq_edge.end())
			//{
			//	if (iter3->get_trav_cost() == DBL_MAX) // this non-required edge is no longer linked
			//	{
			//		this_nonreq_edge.erase(iter3);
			//		this_nonreq_edge_num --;
			//	}
			//	else // this non-required edge is still linked
			//	{
			//		step ++;
			//	}

			//	iter3 = this_nonreq_edge.begin()+step;
			//}

			//// if a required edge is linked but no longer required, move it to the non-required edge. If it is neither required nor linked, remove it from the graph

			//step = 0;
			//iter3 = this_req_edge.begin()+step;
			//while (iter3 != this_req_edge.end())
			//{
			//	if (iter3->get_trav_cost() == DBL_MAX) // this required edge is neither required nor linked
			//	{
			//		this_req_edge.erase(iter3);
			//		this_req_edge_num --;
			//	}
			//	else if (iter3->get_demand() == 0) // this required edge is linked but no longer required
			//	{
			//		//link tmp_link(iter3->get_head(), iter3->get_tail(), iter3->get_trav_cost(), 0, 0, 0);
			//		this_nonreq_edge.push_back(*iter3);
			//		this_req_edge.erase(iter3);
			//		this_nonreq_edge_num ++;
			//		this_req_edge_num --;
			//	}
			//	else // this required edge is still required and linked
			//	{
			//		step ++;
			//	}

			//	iter3 = this_req_edge.begin()+step;
			//}

			// get the total serving cost of this instance

			for (vector<link>::iterator iter2 = this_req_edge.begin(); iter2 != this_req_edge.end(); iter2++)
			{
				if (iter2->get_demand() == 0)
					continue;

				total_serv_cost += iter2->get_serv_cost();
			}

			// check if the uncertain instance is valid, i.e., if the graph is connected

			multi_array<double, 2> this_trav_cost_mtx(extents[vertex_num][vertex_num]);
			multi_array<double, 2> this_dist_mtx(extents[vertex_num][vertex_num]);
			multi_array<double, 2> this_sp_succ(extents[vertex_num][vertex_num]);

			for (int i = 0; i < vertex_num; i++)
			{
				for (int j = 0; j < vertex_num; j++)
				{
					if (j == i)
						continue;

					this_trav_cost_mtx[i][j] = DBL_MAX;
				}
			}

			for (vector<link>::iterator iter2 = this_req_edge.begin(); iter2 != this_req_edge.end(); iter2++)
			{
				if (iter2->get_trav_cost() == DBL_MAX)
					continue;

				this_trav_cost_mtx[iter2->get_head()-1][iter2->get_tail()-1] = iter2->get_trav_cost();
				this_trav_cost_mtx[iter2->get_tail()-1][iter2->get_head()-1] = iter2->get_trav_cost();
			}

			for (vector<link>::iterator iter2 = this_nonreq_edge.begin(); iter2 != this_nonreq_edge.end(); iter2++)
			{
				if (iter2->get_trav_cost() == DBL_MAX)
					continue;

				this_trav_cost_mtx[iter2->get_head()-1][iter2->get_tail()-1] = iter2->get_trav_cost();
				this_trav_cost_mtx[iter2->get_tail()-1][iter2->get_head()-1] = iter2->get_trav_cost();
			}

			dijkstra(this_trav_cost_mtx, this_dist_mtx, this_sp_succ, vertex_num);

			valid = 1;
			for (int i = 0; i < vertex_num; i++)
			{
				for (int j = 0; j < vertex_num; j++)
				{
					if (this_dist_mtx[i][j] == DBL_MAX)
					{
						valid = 0;
						break;
					}
				}
			}

			if (!valid)
				cout << valid << endl;
		}

		for (vector<link>::iterator iter2 = this_req_edge.begin(); iter2 != this_req_edge.end(); iter2++)
		{
			if (iter2->get_trav_cost() == DBL_MAX)
				iter2->set_trav_cost(-1);
		}

		for (vector<link>::iterator iter2 = this_nonreq_edge.begin(); iter2 != this_nonreq_edge.end(); iter2++)
		{
			if (iter2->get_trav_cost() == DBL_MAX)
				iter2->set_trav_cost(-1);
		}
		
		// generate the corresponding output file

		ofstream outfile(iter->c_str());
		string line;

		infile.open(static_instance.c_str(), ifstream::in);
		getline(infile, line); // get the line of NOMBRE
		outfile << line << endl;
		getline(infile, line); // get the line of COMENTARIO
		outfile << " COMENTARIO : unknown. (cota superior)" << endl;
		getline(infile, line); // get the line of VERTICES
		outfile << line << endl;
		getline(infile, line); // get the line of ARISTAS_REQ
		outfile << " ARISTAS_REQ : " << this_req_edge_num << endl;
		getline(infile, line); // get the line of ARISTAS_NOREQ
		outfile << " ARISTAS_NOREQ : " << this_nonreq_edge_num << endl;
		getline(infile, line); // get the line of VEHICULOS
		outfile << line << endl;
		getline(infile, line);  // get the line of CAPACIDAD
		outfile << line << endl;
		getline(infile, line);  // get the line of TIPO_COSTES_ARISTAS
		outfile << line << endl;
		getline(infile, line); // get the line of COSTE_TOTAL_REQ
		outfile << " COSTE_TOTAL_REQ : " << total_serv_cost << endl;
		getline(infile, line); // get the line of LISTA_ARISTAS_REQ
		outfile << line << endl;

		for (int i = 0; i < req_edge_num; i++)
		{
			getline(infile, line);
		}

		for (vector<link>::iterator iter2 = this_req_edge.begin(); iter2 != this_req_edge.end(); iter2++)
		{
			outfile << " ( " << iter2->get_head() << ", " << iter2->get_tail() << ")   tcoste " << iter2->get_trav_cost();
			outfile << "   scoste " << iter2->get_serv_cost() << "   demanda " << iter2->get_demand() << endl;
		}

		if (nonreq_edge_num > 0)
		{
			getline(infile, line); // get the line of LISTA_ARISTAS_NOREQ
			outfile << line << endl;
			for (int i = 0; i < nonreq_edge_num; i++)
			{
				getline(infile, line);
			}

			for (vector<link>::iterator iter2 = this_nonreq_edge.begin(); iter2 != this_nonreq_edge.end(); iter2++)
			{
				outfile << " ( " << iter2->get_head() << ", " << iter2->get_tail() << ")   coste " << iter2->get_trav_cost() << endl;
			}
		}
		
		getline(infile, line); // get the line of DEPOSITO
		outfile << line << endl;

		outfile.close();
		infile.close();
	}
	}
	
	system("pause");
	return 0;
}